import inspect
import dataclass_factory
import json
import base64
from dataclasses import dataclass, field
from typing import (
    Any, Awaitable, Callable, Coroutine,
    Dict, List, Protocol, TypeVar, TypedDict, Union, final
)

from .async_utils import before_first_call


class Context(Protocol):
    function_name: str
    function_version: str
    memory_limit_in_mb: int
    request_id: str
    token: Any

    def get_remaining_time_in_millis(self) -> float: ...


@final
class Identity(TypedDict):
    sourceIp: str
    userAgent: str


@final
class RequestContext(TypedDict):
    identity: Identity
    httpMethod: str
    requestId: str
    requestTime: str  # CLF
    requestTimeEpoch: int


@final
class Event(TypedDict):
    httpMethod: str
    headers: Dict[str, str]
    multiValueHeaders: Dict[str, List[str]]
    queryStringParameters: Dict[str, str]
    multiValueQueryStringParameters: Dict[str, List[str]]
    requestContext: RequestContext
    body: str
    isBase64Encoded: bool


@dataclass
class YaRequest:
    http_method: str
    headers: Dict[str, List[str]]
    query_params: Dict[str, List[str]]
    ctx: Context
    request_ctx: RequestContext
    body: Union[str, bytes]

    @staticmethod
    def build(event: Event, ctx: Context) -> "YaRequest":
        if event["isBase64Encoded"]:
            body = base64.b64decode(event["body"])
        else:
            body = event["body"]

        return YaRequest(
            http_method=event["httpMethod"],
            headers=event["multiValueHeaders"],
            query_params=event["multiValueQueryStringParameters"],
            ctx=ctx,
            request_ctx=event["requestContext"],
            body=body
        )


@dataclass
class YaResponse:
    status_code: int
    body: Union[bytes, str, Dict[str, Any], List[Any]]
    headers: Dict[str, List[str]] = field(default_factory=dict)

    def __post_init__(self):
        if isinstance(self.body, str):
            self._is_b64_encoded = False
        elif isinstance(self.body, bytes):
            self.body = base64.b64encode(self.body).decode("ascii")
            self._is_b64_encoded = True
        elif isinstance(self.body, (list, dict)):
            self.body = json.dumps(self.body)
            self._is_b64_encoded = False
        else:
            raise TypeError(
                "Expected body to be str, bytes, dict or list, got {!r}".format(type(self.body))
            )

    def to_json(self) -> Dict[str, Any]:
        return {
            "statusCode": self.status_code,
            "multiValueHeaders": self.headers,
            "body": self.body,
            "isBase64Encoded": self._is_b64_encoded,
        }


T = TypeVar("T")

_RawHandler = Callable[[Event, Context], Coroutine[Any, Any, Dict[str, Any]]]
_YaHandler = Union[
    Callable[[YaRequest], Awaitable[YaResponse]],
    Callable[[T], Awaitable[YaResponse]],
    Callable[[T, YaRequest], Awaitable[YaResponse]],
]

_Init =  Callable[[], Awaitable[Any]]


async def _default_init():
    return None


def _parse_request(
    request: YaRequest,
    handler: Callable,
    df: dataclass_factory.Factory
):
    signature = inspect.signature(handler)
    params = list(signature.parameters.values())

    if len(params) not in (1, 2):
        raise AssertionError("Handler should have 1 or 2 parameters, got {}".format(len(params)))

    for param in params:
        if param.annotation in (param.empty, YaRequest):
            yield request
        else:
            body_json = json.loads(request.body)
            yield df.load(body_json, param.annotation)


_default_factory = dataclass_factory.Factory(debug_path=True)


def function(
    init: _Init = _default_init,
    df: dataclass_factory.Factory = _default_factory,
) -> Callable[[_YaHandler[T]], _RawHandler]:
    def _decorator(ya_handler: _YaHandler[T]) -> _RawHandler:
        @before_first_call(init)
        async def handler(event: Event, context: Context) -> Dict[str, Any]:
            req = YaRequest.build(event, context)
            try:
                args = list(_parse_request(req, ya_handler, df))
            except (ValueError, TypeError) as e:
                resp = YaResponse(400, "\n".join(map(str, e.args)))
            else:
                resp = await ya_handler(*args)
            return resp.to_json()
        return handler
    return _decorator



import asyncio
from functools import wraps
from typing import Awaitable, Callable, TypeVar, Union


Fn = TypeVar("Fn", bound=Callable[..., Awaitable])


def before_first_call(init: Callable[[], Awaitable]) -> Callable[[Fn], Fn]:
    def _decorator(fn: Fn) -> Fn:
        initialized: Union[bool, asyncio.Event] = False
        # If another call happens while `init()` is still running, we want to
        # wait for it and not start a new coroutine

        @wraps(fn)
        async def _new_fn(*args, **kwargs):
            nonlocal initialized
            if initialized is False:
                e = initialized = asyncio.Event()
                await init()
                e.set()
                initialized = True
            elif initialized is True:
                pass
            else:
                await initialized.wait()

            return await fn(*args, **kwargs)
        return _new_fn  # type: ignore
    return _decorator



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yalambda']

package_data = \
{'': ['*']}

install_requires = \
['dataclass-factory>=2.11,<3.0']

setup_kwargs = {
    'name': 'yalambda',
    'version': '0.3.0',
    'description': 'Utility library for Yandex.Cloud',
    'long_description': '# `yalambda`\n\nYalambda lets you write Yanex.Cloud Functions with less boilerplate\n\nFeatures:\n- everything is type-annotated, so you\'ll get autocompletion in IDEs\n- base64 de/encoding and other details are handled for you\n\n\n# Echo server example\n\n```py\nfrom yalambda import function, YaRequest, YaResponse\n\n\n@function()\nasync def handler(req: YaRequest) -> YaResponse:\n    return YaResponse(200, req.body)\n```\n\n\n# Automatically parse dataclasses\n```py\nfrom dataclasses import dataclass\nfrom yalambda import function, YaResponse\n\n\n@dataclass\nclass Point:\n    x: float\n    y: float\n\n\n@function()\nasync def handler(point: Point) -> YaResponse:\n    dist = (point.x**2 + point.y**2)**0.5\n    return YaResponse(200, {"distance_to_origin": dist})\n```\n\nWe use the `dataclass-factory` package to parse the JSON request\n\n\n# Access both the dataclass and the request\n\n```py\nfrom dataclasses import dataclass\nfrom yalambda import function, YaRequest, YaResponse\n\n\n@dataclass\nclass Point:\n    x: float\n    y: float\n\n\n@function()\nasync def handler(point: Point, req: YaRequest) -> YaResponse:\n    if req.http_method != "POST":\n        return YaResponse(405, "Only POST requests are allowed")\n\n    dist = (point.x**2 + point.y**2)**0.5\n    return YaResponse(200, {"distance_to_origin": dist})\n```\n\n\n# Initialize something asynchronously on first call\n\n```py\nfrom yalambda import function, YaRequest, YaResponse\n\n\nasync def init():\n    global answer\n    answer = 42\n\n\n@function(init)\nasync def handler(req: YaRequest) -> YaResponse:\n    return YaResponse(200, "Answer:".format(answer))\n```',
    'author': 'decorator-factory',
    'author_email': '42166884+decorator-factory@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/decorator-factory/yalambda',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
